@@ -18,7 +18,7 @@
 */
 
 static char *electricsheep_c_id =
-"@(#) $Id: electricsheep.c,v 1.26 2004/04/13 23:29:33 spotspot Exp $";
+"@(#) $Id: electricsheep.c,v 1.45 2005/06/06 21:51:32 spotspot Exp $";
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -38,6 +38,7 @@ static char *electricsheep_c_id =
 #include <expat.h>
 
 #include "config.h"
+#include "getdate.h"
 
 #ifdef HAVE_SYS_MOUNT_H
 #include <sys/mount.h>
@@ -55,19 +56,29 @@ static char *electricsheep_c_id =
 #define STATFS statfs
 #endif
 
-// should be PACKAGE_DATA_DIR
-char *splash_prefix = "/usr/local/share";
+char *splash_prefix = PACKAGE_DATA_DIR "/electricsheep";
+
+#define uniqueid_len 16
+char uniqueid[uniqueid_len+1];
 
 char *proxy_name = 0;
 char *proxy_user = 0;
 
 char *leave_prefix = NULL;
-int leave_max_megabytes = 300;
+int leave_max_megabytes = -1;
 int min_megabytes = 100;
 int reset_fuse_length = 300;
 
+/* only update cache every 100 sheep =~ 6 minutes. */
+int cache_update_delay = 100;
+
+/* wait this long before going to work */
+int init_delay = 600;
+int init_delay_list = 60;
+int list_freshness = 600;
+
 // ick xxx
-#define max_anims 10000
+#define max_anims 100000
 
 #define copy_buffer_size 64000
 
@@ -81,12 +92,21 @@ int nplays = 0;
 int nloopplays = 0;
 int reset_fuse = 0;
 
+int bracket_begin_id = -1;
+int bracket_end_id = -1;
+
+char *logfile = NULL;
+
+time_t bracket_begin_time = (time_t)(-1);
+time_t bracket_end_time = (time_t)(-1);
+
+
 int nplays_by_id[max_anims];
 
 int nrepeats = 2;
 int nthreads = -1;
 
-char *dream_server = "v2d5.sheepserver.net";
+char *dream_server = "v2d6.sheepserver.net";
 
 // cheeze XXX
 #define max_cp_size 300000
@@ -96,8 +116,7 @@ char *dream_server = "v2d5.sheepserver.net";
 char *play_prog = "mpeg2dec_onroot";
 char *nick_name = "";
 char *url_name = "";
-
-// would be good to figure out xscreensaver/redhat/suse etc too?  how?
+char *vote_prog = "electricsheep-voter";
 char *client_version = "LNX_" VERSION;
 
 char cps_name[PATH_MAX] = "";
@@ -107,11 +126,13 @@ char fifo_name[PATH_MAX] = "";
 
 char curl_cmd[MAXBUF];
 
-int history_size = 30;
+char id_file[MAXBUF];
+
+int history_size = 300;
 int *path_history;
 int last_sheep = -1;
 int nrepeated = 0;
-int max_repeats = 10;
+int max_repeats = 3;
 
 
 // by default xscreensaver starts us at priority 10.
@@ -127,12 +148,11 @@ char *window_id = NULL;
 int on_root = 0;
 int timeout = 401;
 int tryagain = 696;
-int jpeg_quality = 90;
 
 char *hide_stderr = "2> /dev/null";
 
 int display_anim = 1;
-int voting = 0;
+int voting = 1;
 int nobg = 0;
 int parasite = 0;
 int standalone = 0;
@@ -155,6 +175,7 @@ int start_id = -1;
 #define max_url_length 1000
 
 typedef struct {
+    int generation;
     int id;
     int deleted;
     int type;
@@ -268,6 +289,13 @@ mysystem2(char *cmd, char *msg) {
     }
 }
 
+void timestamp() {
+#if 0
+  time_t t;
+  time(&t);
+  puts(ctime(&t));
+#endif
+}
 
 static void
 encode(char *dst, char *src) {
@@ -289,8 +317,9 @@ print_anims(char *name, anim_t *an, int nanims) {
     int i;
     printf("%s=%d\n", name, nanims);
     for (i = 0; i < nanims; i++) {
-	printf("id=%d deleted=%d type=%d ctime=%d size=%d rating=%d "
+	printf("gen=%d id=%d deleted=%d type=%d ctime=%d size=%d rating=%d "
 	       "first=%d last=%d url=%s\n",
+	       an[i].generation,
 	       an[i].id, an[i].deleted, an[i].type, an[i].ctime,
 	       an[i].size, an[i].rating, an[i].first, an[i].last,
 	       an[i].url);
@@ -326,7 +355,7 @@ update_cached_anims(int match_gen) {
     ncached_anims = 0;
     while (e = readdir(d)) {
 	anim_t *an = &cached_anims[ncached_anims];
-	int i;
+	int i, dummy;
 	if (match_gen) {
 	    int old_gen = -1;
 	    if (0 == strncmp("gen", e->d_name, 3)) {
@@ -339,8 +368,9 @@ update_cached_anims(int match_gen) {
 	    }
 	}
 	if (filename_is_xxx(e->d_name)) {
-	    if (3 != sscanf(e->d_name, "%d=%d=%d.xxx",
-			    &an->id, &an->first, &an->last)) {
+	    if (4 != sscanf(e->d_name, "%d=%d=%d=%d.xxx",
+			    &an->generation, &an->id,
+			    &an->first, &an->last)) {
 		continue;
 	    }
 	    an->deleted = 1;
@@ -364,13 +394,23 @@ update_cached_anims(int match_gen) {
 	if (!filename_is_mpg(e->d_name)) continue;
 	sprintf(fbuf, "%s%s", leave_prefix, e->d_name);
 
-	if (3 != sscanf(e->d_name, "%d=%d=%d.mpg",
-			&an->id, &an->first, &an->last)) {
+	if (4 != sscanf(e->d_name, "%d=%d=%d=%d.mpg",
+			&an->generation, &an->id, &an->first, &an->last)) {
 	    continue;
 	}
 	    
 	if (-1 == stat(fbuf, &sbuf)) continue;
 
+	if ((bracket_begin_id != -1 && an->id < bracket_begin_id) ||
+	    (bracket_end_id != -1 && an->id > bracket_end_id))
+	  continue;
+
+	if ((bracket_begin_time != (time_t)(-1) &&
+	     sbuf.st_ctime < bracket_begin_time) ||
+	    (bracket_end_time != (time_t)(-1) &&
+	     sbuf.st_ctime > bracket_end_time))
+	  continue;
+
 	an->rating = 0;
 	for (i = 0; i < nserver_anims; i++) {
 	    if (server_anims[i].id == an->id) {
@@ -424,14 +464,14 @@ irandom(int n) {
 
 void
 cached_file_name(char *buf, anim_t *an) {
-    sprintf(buf, "%s%05d=%05d=%05d.mpg",
-	    leave_prefix, an->id, an->first, an->last);
+    sprintf(buf, "%s%05d=%05d=%05d=%05d.mpg",
+	    leave_prefix, an->generation, an->id, an->first, an->last);
 }
 
 void
 deleted_file_name(char *buf, anim_t *an) {
-    sprintf(buf, "%s%05d=%05d=%05d.xxx",
-	    leave_prefix, an->id, an->first, an->last);
+    sprintf(buf, "%s%05d=%05d=%05d=%05d.xxx",
+	    leave_prefix, an->generation, an->id, an->first, an->last);
 }
 
 void
@@ -451,7 +491,7 @@ default_background(char *more) {
 
     if (nobg || (!on_root && !window_id)) return;
     if (more)
-	sprintf(ob, "-merge -at 500,0 %s/electricsheep-%s.tif",
+	sprintf(ob, "-merge -at 500,0 %s-%s.tif",
 		splash_prefix, more);
     else
 	ob[0] = 0;
@@ -462,8 +502,8 @@ default_background(char *more) {
       qbuf[0] = 0;
 
     sprintf(pbuf, "xsetbg %s "
-	    "-border black -at 0,0 %s/electricsheep-splash-0.tif " 
-	    "-merge -center %s/electricsheep-splash-1.tif %s",
+	    "-border black -at 0,0 %s-splash-0.tif " 
+	    "-merge -center %s-splash-1.tif %s",
 	    qbuf, splash_prefix, splash_prefix, ob);
     mysystem2(pbuf, "splash0");
 }
@@ -526,8 +566,6 @@ do_display() {
     int i;
     int idx;
     int niters;
-	
-    update_cached_anims(0);
 
     if (start_id >= 0) {
 	current_anim_id = start_id;
@@ -544,6 +582,7 @@ do_display() {
 	    }
 	}
     }
+
     if (-1 == idx) {
         int n = 0;
         for (i = 0; i < ncached_anims; i++) {
@@ -554,11 +593,13 @@ do_display() {
 	    not_playing();
 	    default_background(0);
 	    sleep(10);
+	    update_cached_anims(0);
 	    return;
 	}
 	idx = irandom(ncached_anims);
-	while (cached_anims[idx].deleted)
+	while (cached_anims[idx].deleted) {
 	    idx = (idx+1)%ncached_anims;
+	}
 	
 	current_anim_id = cached_anims[idx].id;
 	if (debug) printf("found new anim id=%d.\n", current_anim_id);
@@ -612,8 +653,12 @@ do_display() {
 
 	if (debug) {
 	    printf("history =");
-	    for (h = 0; h < history_size; h++)
+	    for (h = 0; h < history_size; h++) {
+	      if (-1 == path_history[h])
+		break;
+	      else
 		printf(" %d", path_history[h]);
+	    }
 	    printf("\n");
 	}
 
@@ -623,7 +668,7 @@ do_display() {
 	    FILE *mpeg_file;
 	    char copy_buf[copy_buffer_size];
 	    size_t n;
-	  
+
 	    mpeg_file = fopen(fbuf, "r");
 	    if (NULL == mpeg_file) {
 		perror(fbuf);
@@ -745,8 +790,9 @@ delete_cached(int size) {
 
 	for (i = 0; i < ncached_anims; i++) {
 	    if (cached_anims[i].deleted) continue;
-	    if (debug) printf("delete cached total=%g id=%d rating=%d\n",
-			  total, cached_anims[i].id, cached_anims[i].rating);
+	    if (0 && debug)
+	      printf("delete cached total=%g id=%d rating=%d\n",
+		     total, cached_anims[i].id, cached_anims[i].rating);
     
 	    total += cached_anims[i].size;
 	    if (!oldest_time ||
@@ -809,13 +855,15 @@ download_anim(int idx) {
 	return;
     }
     if (sbuf.st_size != server_anims[idx].size) {
-	fprintf(stderr, "incomplete sheep id=%d got=%ld want=%ld\n",
-		server_anims[idx].id, (long)sbuf.st_size,
-		(long)server_anims[idx].size);
+
+      if (debug)
+	printf("deleted incomplete sheep id=%d got=%ld want=%ld\n",
+	       server_anims[idx].id, (long)sbuf.st_size,
+	       (long)server_anims[idx].size);
+
 	unlink(tfb);
 	mpg_name[0] = 0;
 	sleep(tryagain);
-	// xxx limited retries
 	return;
     }
 
@@ -835,10 +883,10 @@ get_control_points(char *buf, int buf_size) {
     char pbuf[MAXBUF];
     FILE *cp;
   
-    sprintf(pbuf, "%s --max-time %d --silent --show-error 'http://%s/cgi/get.cgi?"
-	    "nick=%s&url=%s&version=%s' | gunzip -c %s",
-	    curl_cmd, timeout, dream_server, nick_buf,
-	    url_buf, client_version, hide_stderr);
+    sprintf(pbuf, "%s --silent --show-error 'http://%s/cgi/get?"
+	    "n=%s&w=%s&v=%s&u=%s' | gunzip -c %s",
+	    curl_cmd, dream_server, nick_buf,
+	    url_buf, client_version, uniqueid, hide_stderr);
 
     if (debug) printf("get_control_points %s\n", pbuf);
 
@@ -881,10 +929,10 @@ put_image(char *fname, int frame, int anim_id, int gen) {
 	cleanup_and_exit(1);
     }
 
-    sprintf(pbuf, "%s --silent --max-time %d --upload-file %s "
-	    "'http://%s/cgi/put.cgi?frame=%d&id=%d&size=%ld&gen=%d&version=%s'",
-	    curl_cmd, timeout, fname, dream_server, frame, anim_id,
-	    (long)sbuf.st_size, gen, client_version);
+    sprintf(pbuf, "%s --silent --upload-file %s "
+	    "'http://%s/cgi/put?f=%d&id=%d&s=%ld&g=%d&v=%s&u=%s'",
+	    curl_cmd, fname, dream_server, frame, anim_id,
+	    (long)sbuf.st_size, gen, client_version, uniqueid);
     if (debug) printf("about to put %s\n", pbuf);
     mysystem(pbuf, "put image");
 }
@@ -945,6 +993,38 @@ character_handler(void *userData, const XML_Char *s, int len) {
     }
 }
 
+void get_args(char *args, char *fname) {
+  FILE *f;
+  int n, i;
+  char lbuf[MAXBUF];
+  f = fopen(fname, "r");
+  if (NULL == f) {
+    perror(fname);
+    cleanup_and_exit(1);
+  }
+  lbuf[0] = 0;
+  fgets(lbuf, MAXBUF, f);
+  fclose(f);
+  n = strlen(lbuf);
+  args[0] = 0;
+  for (i = 0; i < n; i++) {
+    if (0 == strncmp("args=", lbuf+i, 5)) {
+      char *end, *beg = strchr(lbuf+i, '"');
+      if (NULL == beg) {
+	fprintf(stderr, "missing begin quote after args: %s", lbuf);
+	cleanup_and_exit(1);
+      }
+      end = strchr(beg+1, '"');
+      if (NULL == end) {
+	fprintf(stderr, "missing end quote after args: %s", lbuf);
+	cleanup_and_exit(1);
+      }
+      strncpy(args, beg+1, end-beg-1);
+      break;
+    }
+  }
+}
+
 void
 do_render() {
     char anim_text[max_cp_size];
@@ -1007,8 +1087,10 @@ do_render() {
     {
 	char b[MAXBUF];
 	char strips[MAXBUF];
-	sprintf(b, "nice -%d env in=%s time=%d out=%s quality=%d anim-flame",
-		nice_level, cps_name, get_frame, jpg_name, jpeg_quality);
+	char args[MAXBUF];
+	get_args(args, cps_name);
+	sprintf(b, "nice -n %d env in=%s time=%d out=%s %s flam3-animate",
+		nice_level, cps_name, get_frame, jpg_name, args);
 	if (debug) printf("about to render %s\n", b);
 	if (mysystem(b, "render")) {
 	    if (debug) printf("render failed, trying again later\n");
@@ -1042,76 +1124,6 @@ do_render() {
     cps_name[0] = 0;
 }
 
-/* runs voting, passes votes from xscreensaver to server.
-   reads from a named pipe ~/.sheep/vote-fifo */
-void
-do_ui() {
-    char vf_name[MAXBUF];
-    char gb[MAXBUF];
-    FILE *vf;
-
-    sprintf(fifo_name, "%svote-fifo", leave_prefix);
-    unlink(fifo_name);
-	
-    if (-1 == mkfifo(fifo_name, S_IRWXU)) {
-	fprintf(stderr, "error making fifo.\n");
-	perror(fifo_name);
-	cleanup_and_exit(1);
-    }
-
-    vf = fopen(fifo_name, "r");
-    if (NULL == vf) {
-	fprintf(stderr, "error opening fifo.\n");
-	perror(fifo_name);
-	cleanup_and_exit(1);
-    }
-
-    while (NULL != fgets(gb, MAXBUF, vf)) {
-	char pbuf[MAXBUF];
-	char idfn[MAXBUF];
-	int id, vote;
-	FILE *idf;
-	vote = atoi(gb);
-	sprintf(idfn, "%sid", leave_prefix);
-	idf = fopen(idfn, "r");
-	if (NULL == idf) {
-	    fprintf(stderr, "unable to open current id file.\n");
-	    perror(idfn);
-	    continue;
-	}
-	if (NULL == fgets(gb, MAXBUF, idf)) {
-	    fprintf(stderr, "unable to read current id file.\n");
-	    perror(idfn);
-	    continue;
-	}
-	fclose(idf);
-
-	if (0 == strcmp(gb, "none\n")) {
-	    fprintf(stderr,
-		    "ignored because there is nothing to vote for.\n");
-	    continue;
-	}
-
-	id = atoi(gb);
-
-	if (vote == 0) {
-	    if (debug) printf("not voting zero\n");
-	} else {
-	    sprintf(pbuf, "%s --max-time %d --silent --show-error "
-		    "'http://%s/cgi/vote.cgi?id=%d&vote=%d'",
-		    curl_cmd, timeout, dream_server, id, vote);
-	    if (debug) printf("voting %s\n", pbuf);
-	    mysystem2(pbuf, "curl vote");
-	    default_background((vote > 0) ? "smile" : "frown");
-	    sleep(3);
-	    default_background(0);
-	}
-    }
-
-    fclose(vf);
-    unlink(fifo_name);
-}
-
 void
 list_start_element(void *userData, const char *name, const char **atts) {
     int i = 0;
@@ -1136,6 +1148,7 @@ list_start_element(void *userData, const char *name, const char **atts) {
 	    cleanup_and_exit(1);
 	}
 	memset(an, 0, sizeof(an));
+	an->generation = generation;
 	while (atts[i]) {
 	    const char *a = atts[i+1];
 	    if (!strcmp(atts[i], "id")) {
@@ -1185,17 +1198,23 @@ list_start_element(void *userData, const char *name, const char **atts) {
     } 
 }
 
-/* download anims, delete old anims, update cache */
-void
-do_download() {
+time_t server_anims_timestamp = -1;
+
+void update_server_anims() {
     char pbuf[MAXBUF];
-    char state[MAXBUF];
     FILE *lf;
     int done;
-    int i, j;
-    int best_rating;
-    time_t best_ctime = 0;
-    int best_anim = -1;
+
+    if (-1 == server_anims_timestamp)
+      server_anims_timestamp = time(0);
+    else {
+      time_t now = time(0);
+      if ((now - server_anims_timestamp) < list_freshness) {
+	if (debug) printf("skipping cgi/list\n");
+	return;
+      }
+      server_anims_timestamp = now;
+    }
 	
     XML_Parser parser = XML_ParserCreate(NULL);
     XML_SetElementHandler(parser, list_start_element, get_end_element);
@@ -1203,9 +1222,9 @@ do_download() {
     in_message = 0;
     server_error = 0;
     
-    sprintf(pbuf, "%s --max-time %d --silent --show-error 'http://%s/cgi/list.cgi'"
-	    "| gunzip -c %s", curl_cmd, timeout, dream_server,
-	    hide_stderr);
+    sprintf(pbuf, "%s --silent --show-error 'http://%s/cgi/list?v=%s&u=%s'"
+	    "| gunzip -c %s", curl_cmd, dream_server,
+	    client_version, uniqueid, hide_stderr);
 
     if (debug) printf("list %s\n", pbuf);
 
@@ -1252,6 +1271,21 @@ do_download() {
 	sleep(tryagain);
 	return;
     }
+}
+
+
+
+/* download anims, delete old anims, update cache */
+void
+do_download() {
+    int i, j;
+    int best_rating;
+    time_t best_ctime = 0;
+    int best_anim = -1;
+
+    timestamp();
+
+    update_server_anims();
 
     if (debug > 1)
 	print_anims("nserver_anims", server_anims, nserver_anims);
@@ -1302,6 +1336,7 @@ make_download_process() {
 	setproctitle("download");
 #endif
 	thread_name = "download";
+	sleep(init_delay_list);
 	while (1) {
 	    do_download();
 	}
@@ -1309,28 +1344,34 @@ make_download_process() {
     /* parent returns */
 }
 
+
 void
 make_ui_process() {
-    
+
+  if (!on_root && !window_id)
+    return;
+
     if (-1 == (ui_pid = fork()))
 	perror("ui fork");
     else if (0 == ui_pid) {
-	/* child */
-#ifdef HAVE_SETPROCTITLE
-	setproctitle("ui");
-#endif
-	thread_name = "ui";
-	while (1) {
-	    do_ui();
-	}
+	char vote_format[MAXBUF];
+
+	execlp(vote_prog, vote_prog,
+	       on_root ? "-3" : window_id,
+	       nick_buf, url_buf, id_file, dream_server,
+	       leave_prefix, uniqueid, NULL);
+	perror("exec vote_prog");
     }
     /* parent returns */
 }
 
+
 void
 make_render_process() {
     pid_t p;
 
+    sleep(init_delay);
+
     while (nthreads-- > 1) { 
 	if (-1 == (p = fork()))
 	    perror("render fork");
@@ -1356,6 +1397,8 @@ make_render_process() {
 void
 make_display_process()  {
     int h;
+    int cnt;
+    pid_t generator_pid;
     for (h = 0; h < history_size; h++)
 	path_history[h] = -1;
     
@@ -1378,11 +1421,28 @@ make_display_process()  {
 	    /* child */
 	    sprintf(fps, "%d", frame_rate);
 	    if (use_mplayer) {
+	      int devnull = open("/dev/null", O_WRONLY);
+	      if (-1 == devnull) {
+		perror("/dev/null");
+		cleanup_and_exit(1);
+	      }
+	      if (-1 == dup2(devnull, STDOUT_FILENO)) {
+		perror("dup2a");
+		cleanup_and_exit(1);
+	      }
+	      if (-1 == dup2(devnull, STDERR_FILENO)) {
+		perror("dup2b");
+		cleanup_and_exit(1);
+	      }
 		argv[c++] = "mplayer";
+		argv[c++] = "-demuxer";
+		argv[c++] = "1";
+		argv[c++] = "-really-quiet";
 		argv[c++] = "-fps";
 		argv[c++] = fps;
 		if (display_zoomed) {
 		    argv[c++] = "-zoom";
+		    argv[c++] = "-fs";
 		}
 		if (on_root) {
 		    argv[c++] = "-rootwin";
@@ -1421,8 +1481,6 @@ make_display_process()  {
 		perror("decoder child dup2 1");
 		cleanup_and_exit(1);
 	    }
-	    close(STDOUT_FILENO);
-	    close(STDERR_FILENO);
 	    execvp(play_prog, argv);
 	    perror("exec play_prog");
 	} else {
@@ -1435,10 +1493,28 @@ make_display_process()  {
 		perror("fdopen 1");
 		cleanup_and_exit(1);
 	    }
-	    while (1) {
+
+	    if ((generator_pid=fork())==0) {
+	      
+	      update_cached_anims(0);
+	      cnt = 0;
+	      while (1) {
+		timestamp();
 		do_display();
+		if (!standalone && ((cnt++%cache_update_delay) == 0)) {
+		  update_cached_anims(0);
+		}
+	      }
+	      /* child never gets here */
+	    } else {
+	      /* wait end of decoder pid (kill window, e.g. screensaver stop on KDE) */
+	      if (-1 == waitpid(decoder_pid, 0, 0)) {
+		perror("waitpid downloader_pid");
+		exit(1);
+	      }	    
+	      cleanup_and_exit(1);
 	    }
-	    /* child never gets here */
+	    
 	}
     }
     /* parent returns */
@@ -1455,7 +1531,7 @@ logo_only() {
   else
     qbuf[0] = 0;
   
-  sprintf(pbuf, "xsetbg %s -at 0,0 %s/electricsheep-splash-1.tif",
+  sprintf(pbuf, "xsetbg %s -at 0,0 %s-splash-1.tif",
 	  qbuf, splash_prefix);
   mysystem2(pbuf, "logo");
   while (1) sleep(60);
@@ -1486,6 +1562,25 @@ auto_nthreads() {
 #endif
 }
 
+void parse_bracket(char *arg, int *bracket_id, time_t *bracket_time) {
+  int i, id, n;
+  if (0 == arg || 0 == arg[0]) return;
+  n = strlen(arg);
+  for (i = 0; i < n; i++) {
+    if (!isdigit(arg[i])) {
+      if (-1 == (*bracket_time = get_date(arg, NULL))) {
+	fprintf(stderr, "warning: badly formated date ignored: %s\n", arg);
+      }
+      return;
+    }
+  }
+  if (1 == sscanf(arg, "%d", &id)) {
+    *bracket_id = id;
+    return;
+  }
+  fprintf(stderr, "bad conversion of %s\n", arg);
+}
+
 #define iarg(oname, vname) \
 if (!strcmp(oname, o)) { \
       if (*argc > 2) \
@@ -1522,11 +1617,10 @@ char *help_string =
 "\n"
 "--nick name (credit frames to this name on the server, default is none)\n"
 "--url url (in server credit, link name to this url, default is none)\n"
-"--nthreads N (number of rendering threads, default is same\n"
-"              as the number of CPUs)\n"
+"--nthreads N (number of rendering threads, default is 1)\n"
 "--frame-rate N (frames/second)\n"
 "--timeout N (seconds, default is 401)\n"
-"--tryagain N (seconds between retries to server, default is 396)\n"
+"--tryagain N (seconds between retries to server, default is 696)\n"
 "--server host/path (a hostname, possibly with a path, no leading\n"
 "              \"http://\", default is %s)\n" 
 "--display-anim 0/1 (invisibility if 0, default 1)\n"
@@ -1535,7 +1629,7 @@ char *help_string =
 "--reset-fuse N (maximum number of transitions before resetting\n"
 "               to avoid loops, -1 means never, default is 300)\n"
 "--max-megabytes N (maximum disk space used to save anims in megabytes,\n"
-"               default is 100, 0 means no limit)\n"
+"               default is 1000, 0 means no limit)\n"
 "--min-megabytes N (minimum disk space to leave free in megabytes,\n"
 "               default is 100, 0 means no limit)\n"
 "--nice n (priority adjustment for render process, default 10)\n"
@@ -1552,11 +1646,17 @@ char *help_string =
 "--zoom 0/1 (zoom to fullscreen, default 0)\n"
 "--player exec_name (use the specified decoder)\n"
 "--history N (length of history to keep in sheep, default 30)\n"
+"--bracket-begin id/date (play no sheep before this one or this time)\n"
+"--bracket-end id/date (play no sheep after this one or this time)\n"
+"--data-dir dir (directory to find splash images and other data files)\n"
+"--logfile file (file to write the log instead of stdout)\n"
 ;
 
 void
 flags_init(int *argc, char ***argv) {
     char *arg0 = (*argv)[0];
+    char *bracket_begin;
+    char *bracket_end;
     while (*argc > 1 &&
 	   (*argv)[1][0] == '-') {
 	char *o = (*argv)[1];
@@ -1584,6 +1684,7 @@ flags_init(int *argc, char ***argv) {
 	else iarg("--debug", debug)
 	else sarg("--player", play_prog)
 	else sarg("--proxy", proxy_name)
+	else sarg("--data-dir", splash_prefix)
 	else iarg("--start-sheep", start_id)
 	else sarg("--proxy-user", proxy_user)
 	else sarg("-window-id", window_id)
@@ -1593,6 +1694,9 @@ flags_init(int *argc, char ***argv) {
 	else iarg("--mplayer",use_mplayer)
 	else iarg("--zoom",display_zoomed)
 	else iarg("--history",history_size)
+	else sarg("--bracket-begin",bracket_begin)
+	else sarg("--bracket-end",bracket_end)
+	else sarg("--logfile",logfile)
 	else {
 	    fprintf(stderr, "bad option: %s, try --help\n", o);
 	    exit(1);
@@ -1600,6 +1704,9 @@ flags_init(int *argc, char ***argv) {
     }
     (*argv)[0] = arg0;
 
+    parse_bracket(bracket_begin, &bracket_begin_id, &bracket_begin_time);
+    parse_bracket(bracket_end, &bracket_end_id, &bracket_end_time);
+
     if (nice_level < -20) {
 	fprintf(stderr, "nice level must be -20 or greater, not %d.\n",
 		nice_level);
@@ -1627,7 +1734,9 @@ flags_init(int *argc, char ***argv) {
     path_history = malloc(sizeof(int) * history_size);
 
     if (-1 == nthreads) {
-	auto_nthreads();
+	nthreads = 1;
+    } else if (0 > nthreads) {
+        auto_nthreads();
     }
 
     if (debug) {
@@ -1650,6 +1759,8 @@ flags_init(int *argc, char ***argv) {
 	leave_prefix = strdup(b);
     }
 
+    sprintf(id_file, "%sid", leave_prefix);
+
     if (1) {
 	char b[MAXBUF];
 	sprintf(b, "mkdir -p %s", leave_prefix);
@@ -1713,6 +1824,32 @@ do_lock() {
     /* leave the file open & locked until our process terminates */
 }
 
+/* 64 random bits encoded as ascii */
+void set_uniqueid() {
+  static char *rdevice = "/dev/urandom";
+  long d[2];
+  int i, rfd;
+  struct timeval tv;
+
+  if (debug) printf("setting unique id.\n");
+  rfd = open(rdevice, 0);
+  if (-1 == rfd) {
+    perror(rdevice);
+    exit(1);
+  }
+  if (8 != read(rfd, (void *) d, 8)) {
+    perror(rdevice);
+    exit(1);
+  }
+  if (-1 == gettimeofday(&tv, NULL)) {
+    perror("gettimeofday");
+    exit(1);
+  }
+  d[0] ^= tv.tv_sec;
+  d[1] ^= tv.tv_usec;
+  sprintf(uniqueid, "%08X%08X", d[0], d[1]);
+}
+
 int
 main(int argc, char **argv) {
     int n = 0;
@@ -1737,22 +1874,84 @@ main(int argc, char **argv) {
 	for (i = 0; i < max_anims; i++)
 	    nplays_by_id[i] = 0;
     }
-	
+
     flags_init(&argc, &argv);
 
+    if (1) {
+      /* read rc file */
+      FILE *frc;
+      char tbuf[MAXBUF];
+      char pbuf[MAXBUF];
+      int changed = 0;
+
+      sprintf(pbuf, "%src", leave_prefix);
+      frc = fopen(pbuf, "rb");
+      if (NULL != frc) {
+	int tlmm;
+	if (1 == fscanf(frc, "%d\n", &tlmm)) {
+	  if (-1 == leave_max_megabytes) {
+	    leave_max_megabytes = tlmm;
+	  } else if (leave_max_megabytes != tlmm) {
+	    changed = 1;
+	  }
+	} else {
+	  fprintf(stderr, "warning: could not parse line 1 of %s. fixed.\n",
+		  pbuf);
+	  changed = 1;
+	}
+	uniqueid[0] = 0;
+	if (NULL == fgets(uniqueid, uniqueid_len+1, frc) ||
+	    strlen(uniqueid)!=uniqueid_len) {
+	  fprintf(stderr, "warning: could not parse line 2 of %s. fixed.\n",
+		  pbuf);
+	  set_uniqueid();
+	  changed = 1;
+	}
+	fclose(frc);
+      } else {
+	set_uniqueid();
+	changed = 1;
+      }
+
+      if (0 > leave_max_megabytes) {
+	leave_max_megabytes = 1000;
+	changed = 1;
+      }
+
+      /* write rc file */
+      if (changed) {
+	sprintf(tbuf, "%src.tmp", leave_prefix);
+	if (frc = fopen(tbuf, "wb")) {
+	  fprintf(frc, "%d\n%s\n", leave_max_megabytes, uniqueid);
+	  fclose(frc);
+	  if (-1 == rename(tbuf, pbuf)) {
+	    perror(pbuf);
+	  }
+	} else {
+	  perror(tbuf);
+	}	  
+      }
+    }
+
+
+    if (logfile) {
+      freopen(logfile, "ab", stdout);
+      freopen(logfile, "ab", stderr);
+    }
+
     do_lock();
 
     reset_fuse = reset_fuse_length;
 
     if (proxy_name) {
-	sprintf(curl_cmd, "nice -%d curl --proxy %s",
+	sprintf(curl_cmd, "nice -n %d curl --proxy %s",
 		nice_level, proxy_name);
 	if (proxy_user) {
 	  strcat(curl_cmd, " --proxy-user ");
 	  strcat(curl_cmd, proxy_user);
 	}
     } else
-	sprintf(curl_cmd, "nice -%d curl", nice_level);
+	sprintf(curl_cmd, "nice -n %d curl", nice_level);
 
     srandom(time(0));
 
